/**
 * @file  Channel4Protocol.cpp
 * @brief Implementierung der Klasse CChannel4Protocol.
 */
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Channel4Protocol.h"
#include <stdio.h>

// ##################################################################
/**
 * @brief Initialisieren der Membervariablen.
 *
 * @param channelID Nummer der Kanalklasse fuer z.B. Trace.
 * @param deviceID Busnummer der Kanalklasse.
 */
void CChannel4Protocol::InitClass(short channelID, short deviceID)
{
	Trace("%s(%i):INFO(3): CChannel4Protocol::InitClass [ %i, %i ]\n", __FILE__, __LINE__, channelID, deviceID);

	// Umgebung zuruecksetzen
	Reset();

	// DeviceID der Klasse zuweisen
	m_deviceID = deviceID;

	// Zustand des Kanalobjekts
	m_activeState = is_free;

	// Puffer zur schnellen Uebertragung
	memset(m_tmp, 0, sizeof(m_tmp));
	m_tmpPos = m_tmp;

	// Die Umgebung des verwendeten CRC32 initialisieren
	InitCRC32Table();

	// Membervariablen fuer Paketauswertung initialisieren
	m_packetStart = FALSE;
	m_skipCount = 0;
	m_packetNumber = 0;
	m_lastFlags = 0;

	m_clocksTx = m_clockTx = 0;
	m_clocksRx = m_clockRx = 0;
	
	m_transmittedPackets = 0;
	m_transmittedSize = 0;
	m_transmittedSizeData = 0;

	m_receivedPackets = 0;
	m_receivedPacketsReject = 0;
	m_receivedSize = 0;
	m_receivedSizeData = 0;

	m_qs = 0;

	// Callbacks intialisieren
	m_callBack1 = NULL;
	m_callBack2 = NULL;

	// Puffer fuer Senden, Empfang
	memset(m_transmit, 0, sizeof(m_transmit));
	memset(m_receive, 0, sizeof(m_receive));

	// Hilfsmember initialisieren
	m_lastModule   = 0;
	m_lastFunction = 0;
	m_result = 0;

	// Zufallsgenerator starten
	srand((unsigned int)time(NULL));

	// Verschluesselung aktivieren
//	DF4Init(TRUE);

	// Kommunikationsbasis an TimeboyStation weitergeben
	m_timeboyStation.SetChannelBase(GetChannelBase());
}

// ##################################################################
/**
 * @brief Destruktor.
 *
 */
CChannel4Protocol::~CChannel4Protocol() 
{
}


//#########################################################
//####################### HELPER ##########################
//#########################################################

// ##################################################################
/**
 * @brief Initialisiert die grundlegend benoetigten Membervariablen.
 */
void CChannel4Protocol::Reset()
{
	Trace("%s(%i):INFO(3): CChannel4Protocol::Reset\n", __FILE__, __LINE__);

	m_result = IV_ERROR_SUCCESS;
	ResetError();
	memset(m_data, 0, sizeof(m_data));
}

// ##################################################################
/**
 * @brief Setzen einer auszufuehrenden Call-Back Funktion.
 */
void CChannel4Protocol::SetCBFunction(int (*pFunction)(void))
{	
	Trace("%s(%i):INFO(3): CChannel4Protocol::SetCBFunction [ 0x%08X ]\n", __FILE__, __LINE__, pFunction);

	m_callBack1 = pFunction;
}

// ##################################################################
/**
 * @brief Setzen einer auszufuehrenden Call-Back Funktion mit Parametern.
 */
void CChannel4Protocol::SetCBFunctionWithParams(int (*pFunction)(unsigned int, unsigned int))
{
	Trace("%s(%i):INFO(3): CChannel4Protocol::SetCBFunctionWithParams [ 0x%08X ]\n", __FILE__, __LINE__, pFunction);

	m_callBack2 = pFunction;
}

int CChannel4Protocol::BinToUIntMSBFirst(unsigned char *Buf)
{
	return int(Buf[0] << 8) | int(Buf[1]);
}

void CChannel4Protocol::ULongToBinMSBFirst(unsigned char *Buf, unsigned int Val)
{
	Buf[3] = (unsigned char)(Val & 0xFF);
	Val >>= 8;
	Buf[2] = (unsigned char)(Val & 0xFF);
	Val >>= 8;
	Buf[1] = (unsigned char)(Val & 0xFF);
	Val >>= 8;
	Buf[0] = (unsigned char)(Val & 0xFF);
}

void CChannel4Protocol::UIntToBinMSBFirst(unsigned char *Buf, unsigned int Val)
{
	Buf[1] = (unsigned char)(Val & 0xFF);
	Val >>= 8;
	Buf[0] = (unsigned char)(Val & 0xFF);
}

unsigned int CChannel4Protocol::BinToULongMSBFirst(unsigned char *Dat)
{
	unsigned int Val;

	Val  = Dat[0]; Val <<= 8;
	Val |= Dat[1]; Val <<= 8;
	Val |= Dat[2]; Val <<= 8;
	Val |= Dat[3];
	return Val;
}


//#########################################################
//######################### CRC ###########################
//#########################################################

// ##################################################################
/**
 * @brief Reflection is a requirement for the official CRC-32 standard.
 * You can create CRCs without it, but they won't conform to the standard.
 */
unsigned int CChannel4Protocol::Reflect(unsigned int ref, char ch)
{// Used only by InitCRC32Table()
	Trace("%s(%i):INFO(3): CChannel4Protocol::Reflect [ %i, %i ]\n", __FILE__, __LINE__, ref, ch);

	unsigned int val = 0;

	// Swap bit 0 for bit 7
	// bit 1 for bit 6, etc.
	int i;
	for(i=1; i < (ch + 1); i++)
	{
		if (ref & 1)
		{
			val |= 1 << (ch - i);
		}
		ref >>= 1;
	}

	return val;
}

// ##################################################################
/**
 * @brief Call this func only once to initialize the CRC table.
 */
void CChannel4Protocol::InitCRC32Table() 
{// Called by Init
	Trace("%s(%i):INFO(3): CChannel4Protocol::InitCRC32Table\n", __FILE__, __LINE__);

	// This is the official polynomial used by CRC-32 
	// in PKZip, WinZip and Ethernet. 
	unsigned int ulPolynomial = 0x04c11db7;

	// 256 values representing ASCII character codes.
	int i, j;
	for(i=0; i <= 0xFF; i++)
	{
		m_crc32_table[i] = Reflect(i, 8) << 24;
		for (j=0; j < 8; j++)
		{
			m_crc32_table[i] = (m_crc32_table[i] << 1) ^ (m_crc32_table[i] & (1 << 31) ? ulPolynomial : 0);
		}

		m_crc32_table[i] = Reflect(m_crc32_table[i], 32);
	}
}

// ##################################################################
/**
 * @brief This func uses the crc32_table lookup table to generate a CRC for data.
 */
unsigned int CChannel4Protocol::GetCRC32(unsigned char *data, unsigned int size, unsigned int crc, unsigned char closure)
{
	Trace("%s(%i):INFO(3): CChannel4Protocol::GetCRC32 [ 0x%08X, %i, 0x%08X, %i ]\n", __FILE__, __LINE__, data, size, crc, closure);

	// Be sure to use unsigned variables,
	// because negative values introduce high bits
	// where zero bits are required.
	//unsigned int crc = 0xffffffff;
	int len;
	unsigned char *buffer;

	len = size;
	// Save the text in the buffer.
	buffer = data;
	// Perform the algorithm on each character
	// in the string, using the lookup table values.
	while (len--)
	{
		crc = (crc >> 8) ^ m_crc32_table[(crc & 0xFF) ^ *buffer++];
	}
	// Exclusive OR the result with the beginning val.
	return (closure) ? crc ^ 0xffffffff : crc;
}

// ##################################################################
/**
 * @brief Erzeugt einen 16bit CRC Wert fuer die uebergebenen Daten.
 */
unsigned short CChannel4Protocol::GetCRC16(unsigned char *data, unsigned int size, unsigned short crc)
{
	Trace("%s(%i):INFO(3): CChannel4Protocol::GetCRC16 [ 0x%08X, %i, 0x%04X ]\n", __FILE__, __LINE__, data, size, crc);

	unsigned int i;
	unsigned char j;
  
    for (i=0; i < size; i++)
	{
        crc = crc ^ data[i];

        for (j=0; j < 8; j++)
		{
            if (crc & 0x1)
			{
				crc >>= 1; 
				crc = crc ^ 0xA001;
			}
			else 
			{
				crc >>= 1;
			}
        }
    }

	return crc;
}


//#########################################################
//################ Verschluesselung #######################
//#########################################################

void CChannel4Protocol::DF4LoadCryptKey(void)
{
	unsigned char cryptKey_DF4[32] = 
	{	
		0xc5, 0xc2, 0xc3, 0xc0, 0xc1, 0xd9, 0xdb, 0xf0, 
		0xcd, 0x07, 0xa7, 0xaf, 0xaa, 0x0a, 0x50, 0xae,
		0xc5, 0x4e, 0xc3, 0xd9, 0xaf, 0xd9, 0x09, 0xc1,
		0xa7, 0xef, 0xac, 0xc2, 0x0a, 0xbf, 0xcd, 0xae
	};

	memcpy(m_key, cryptKey_DF4, sizeof(cryptKey_DF4));
}

// ##################################################################
/**
 * @brief Initialisiert die sbox fuer die DF4-Verschlusselungsroutine.
 */
void CChannel4Protocol::DF4Init(int binaryKey)
{
	unsigned i, j;
	unsigned char *kptr = m_key, trans;
	
//	ASSERT(sizeof(m_sbox) == 256);
//	ASSERT(sizeof(m_k) == 256);

	for(i=0; i < 256; i++)
	{		
		m_sbox[i] = (unsigned char)i;
	}

	for(i=0; i < 256; i++)
	{
		m_k[i] = *kptr;
		kptr++;

		if (binaryKey)
		{
			if(kptr == (m_key + sizeof(m_key)))
			{				
				kptr = m_key;
			}
		}
		else
		{
			if (*kptr == 0)
			{
				kptr = m_key;
			}
		}
	}

	for(i=0, j=0; i < 256; i++)
	{
		j = (j + m_sbox[i] + m_k[i]) % 256;
		trans = m_sbox[i];
		m_sbox[i] = m_sbox[j];
		m_sbox[j] = trans;
	}
}

// ##################################################################
/**
 * @brief Ver-/Entschluesselt die uebergebenen Daten.
 */
void CChannel4Protocol::DF4(unsigned char *data, int size)

{
	unsigned i, j, t, pos;
	unsigned char trans;

	// Kopie von sbox, da sie unten manipuliert wuerde.
	memcpy(m_k, m_sbox, sizeof(m_k));

	// Wenn die Angabe einer Laenge fehlt, wird davon ausgegangen,
	// dass mit einer Zeichenkette gearbeitet wird.
	if (size == 0)
	{
		size = (int)strlen((char*)data);
	}

	i = j = pos = 0;
	for(; (size - pos) > 0; pos++)
	{
		i = (i + 1) % 256;
		j = (j + m_k[i]) % 256;
		trans = m_k[i];
		m_k[i] = m_k[j];
		m_k[j] = trans;
		t = (m_k[i] + m_k[j]) % 256;
		data[pos] ^= m_k[t];
	}
}

// ##################################################################
/**
 * @brief Schreiben eines teilaufbereiteten Packets in die Kommunikationsschnittstelle.
 * Der Kopf und Packet-CRC wird vor Packetversand gebildet und zur Packetgroesse dazugerechnet
 * und nach Versand wieder entfernt.
 */
int CChannel4Protocol::TransmitProtocol(unsigned char *buffer, int transmitTrials)
{
	Trace("%s(%i):INFO(3): CChannel4Protocol::TransmitProtocol [ 0x%08X, %i ]\n", __FILE__, __LINE__, buffer, transmitTrials);

	int res = IV_ERROR_SUCCESS;
	int transmittedCount = 0, transmitted, enableMasking = 1, masked = 0;

	// Laengenangabe pruefen
	if (*(unsigned short*)&buffer[4] < IV_PACKET_LENGTH_FRAME &&
		*(unsigned short*)&buffer[4] >= IV_PACKET_LENGTH_MAX) 
	{
		Trace("%s(%i):ERROR(1): CChannel4Protocol::TransmitProtocol [ packet length fail: %i ]\n", __FILE__, __LINE__, -IV_ERROR_PACKETLENGTH);
		return -IV_ERROR_PACKETLENGTH;
	}

	// Temppuffer setzen
	m_tmpPos = m_tmp;

	// Handelt es sich um eine Dockingstation?
	if (IsTimeboySlot(GetDeviceIDRemote()))
	{
		// Keine Maskierung der Paketdaten bei versenden ueber eine Dockingstation.
		enableMasking = 0;
	}

	// CRC - groesse daraufrechnen
	*(unsigned short*)&buffer[4] += sizeof(unsigned short);
	// CRC von Paketkopf einfuegen
	*(unsigned short*)&buffer[8] = GetCRC16(buffer, IV_PACKET_LENGTH_HEAD - sizeof(unsigned short));
	// CRC von Paket anfuegen
	*(unsigned short*)&buffer[*(unsigned short*)&buffer[4] - sizeof(unsigned short)] = GetCRC16(buffer, *(unsigned short*)&buffer[4] - sizeof(unsigned short));
/*
	if (IV_PACKET_MODULE(buffer) != 0) 
	{
		// CRC von Paket flschen
		if (!(rand() % 8)) 
		{
			*(unsigned short*)&buffer[*(unsigned short*)&buffer[4] - sizeof(unsigned short)] += 1;
			Trace("%s(%i):ERROR(1): CChannel4Protocol::TransmitProtocol [ verflsche gesamt CRC ]\n", __FILE__, __LINE__);
		}
	}
*/

	// Datenstrom zeichenweise bearbeiten und besondere Zeichen maskieren
	for (; transmittedCount < *(unsigned short*)&buffer[4]; transmittedCount++) 
	{	
		// Paketzeichen im Datenstrom mit Escapesequenz umschreiben
		if (enableMasking && (transmittedCount > 1) && 
			((buffer[transmittedCount] == IV_PACKET_SIGN_START) ||
			 (buffer[transmittedCount] == IV_PACKET_SIGN_ESCAPE)) ) 
		{
			unsigned char replace[2];
			replace[0] = IV_PACKET_SIGN_ESCAPE;
			// Anzahl maskierte mitzaehlen
			masked++;

			switch (buffer[transmittedCount]) 
			{
			case IV_PACKET_SIGN_START:  replace[1] = 0x01; break;
			case IV_PACKET_SIGN_ESCAPE: replace[1] = 0x00; break;
			}

			// Wuerde Pufferende erreicht?
			if (((m_tmpPos + sizeof(replace)) - m_tmp) > (int)sizeof(m_tmp)) 
			{
				Trace("%s(%i):ERROR(1): CChannel4Protocol::TransmitProtocol [ Puffer overflow: %i ]\n", __FILE__, __LINE__, -IV_ERROR_TRANSMIT);
				return -IV_ERROR_TRANSMIT;
			}

			*m_tmpPos = ((unsigned char*)replace)[0];
			m_tmpPos++; // Auf naechstes Zeichen

			*m_tmpPos = ((unsigned char*)replace)[1];
			m_tmpPos++; // Auf naechstes Zeichen

		} 
		else 
		{
			// Wuerde Pufferende erreicht?
			if (((m_tmpPos + 1) - m_tmp) > (int)sizeof(m_tmp)) 
			{
				Trace("%s(%i):ERROR(1): CChannel4Protocol::TransmitProtocol [ Puffer overflow: %i ]\n", __FILE__, __LINE__, -IV_ERROR_TRANSMIT);
				return -IV_ERROR_TRANSMIT;
			}

			*m_tmpPos = buffer[transmittedCount];
			m_tmpPos++; // Auf naechstes Zeichen
		}
	}

	// Absenden des Gesamtpaketes
	if (transmittedCount == *(unsigned short*)&buffer[4]) 
	{
		m_clockTx = 0;
		// Anfang des Versendens halten
		clock_t clockStart = clock();
		if (IsTimeboySlot(GetDeviceIDRemote()))
		{
			if ((res = m_timeboyStation.WriteBytes(GetDeviceIDRemote(), m_timeboyBaseAddress, 0, m_tmp, *(unsigned short*)&buffer[4] + masked)) != IV_ERROR_SUCCESS)
			{
				Trace("%s(%i):ERROR(1): CChannel4Protocol::TransmitProtocol [ WriteBytes fail with: %i ]\n", __FILE__, __LINE__, res);
				goto label_error;
			}
		}
		else
		{
			if ((res = TransmitData(m_tmp, *(unsigned short*)&buffer[4] + masked, transmitted, transmitTrials)) != IV_ERROR_SUCCESS) 
			{
				Trace("%s(%i):ERROR(1): CChannel4Protocol::TransmitProtocol [ TransmitData fail with: %i ]\n", __FILE__, __LINE__, res);
				goto label_error;
			}
		}

		// Gesamtzeit des Versendevorgangs fuer Statistik halten
		m_clockTx = clock() - clockStart;
		m_clocksTx += m_clockTx;
	}

label_error:
	// CRC - groesse wieder abziehen
	*(unsigned short*)&buffer[4] -= sizeof(unsigned short);
	// CRC von Paketkopf loeschen
	*(unsigned short*)&buffer[8] = 0;
	// CRC von Paket loeschen
	*(unsigned short*)&buffer[*(unsigned short*)&buffer[4]] = 0;

	return res;
}

// ##################################################################
/**
 * @brief Schreiben eines Packets fuer die angegebene Modulfunktion.
 *
 * @param deviceIDRemote Anzusprechende Geraetenummer.
 */
int CChannel4Protocol::TransmitPacket(short deviceIDRemote, unsigned char type, short module, short func, unsigned char *buf, int count, int transmitTrials)
// Aufbereiten und senden eines Anfrage-Pakets
//  deviceIDRemote -> Adresse (Geraetenummer) des Empfaengers
//  type     -> Pakettyp (Anfrage/Antwort +? Error)
//  module   -> Nummer des anzusprechenden Modules
//  func -> Nummer der Funktion im Module
//	buf	     -> Buffer fuer Funktionsparameter.
//	count    -> Anzahl der Funktionsparameter in Byte. Rueckgabewert bei Parameterempfang.
//
//  return -> <0 -> Kommunikationsfehler
//             0 -> senden und empfangen und ausfuehren erfolgreich
//			  >0 -> module-/Funktionsfehler
{
	Trace("%s(%i):INFO(3): CChannel4Protocol::TransmitPacket [ %i, %i, %i, 0x%08X, %i ]\n", __FILE__, __LINE__, deviceIDRemote, module, func, buf, count);

	int err = IV_ERROR_SUCCESS;

	// CallBack-Function vor Packetversant aufrufen.
	if (m_callBack1) 
	{
		if (!m_callBack1()) 
		{	
			Trace("%s(%i):ERROR(1): CChannel4Protocol::TransmitPacket [ CallBack abbruch ]\n", __FILE__, __LINE__);
			return -IV_ERROR_BREAK;
		}
	}

	// Remoteadresse setzen
	SetDeviceIDRemote(deviceIDRemote);

/* 
1.	(  8 Bit): Startzeichen ("~", 0x7E)
2.	(  8 Bit): Empfngeradresse
3.	(  8 Bit): Absenderadresse
4.	(  8 Bit): Steuerinformationslnge
5.	(16 Bit): Paketgesamtlnge
6.	(  8 Bit): Flags (z. B. fr aktive Verschlsselung)
7.	(  8 Bit): Reserve
8.	(16 Bit): CRC ber die ersten 7 Felder
9.	(n*8 Bit): Steuerinformationsdaten (in Bezug auf Feld 5)
10.	(n*8 Bit): Zu transportierende Daten
11.	(16 Bit): CRC ber Felder 1 - 10.
*/
	// 1.
	m_transmit[0] = IV_PACKET_SIGN_START;
	// 2.
	m_transmit[1] = (unsigned char)GetDeviceIDRemote();
	// 3.
	m_transmit[2] = (unsigned char)GetDeviceID();
	// 4.
	m_transmit[3] = IV_PACKET_LENGTH_CONTROLBLOCK;
	// 5.
	*(unsigned short*)&m_transmit[4] = IV_PACKET_LENGTH_HEAD + m_transmit[3] + (unsigned short)sizeof(unsigned char) + (unsigned short)sizeof(unsigned short) * 2 + count;
	// 6.
	m_transmit[6] = 0;
	// Flags fuer Verschluesselung setzen.
	//if (IsKeyOEM2()) {	m_transmit[6] |= IV_PACKET_FLAG_DF4_KEY_OEM2; }
	// 7.
	m_transmit[7] = 0;
	// 8.
	// Wird von Transmit - Funktion erledigt
	// 9.
	// Wird kurz vor Versand erledigt.
	// 10.
	m_transmit[IV_PACKET_LENGTH_HEAD + m_transmit[3]] = IV_PACKET_VERSION_4 | type;
	*(unsigned short*)&m_transmit[IV_PACKET_LENGTH_HEAD + m_transmit[3] + sizeof(unsigned char)]= module;
	*(unsigned short*)&m_transmit[IV_PACKET_LENGTH_HEAD + m_transmit[3] + sizeof(unsigned char) + sizeof(unsigned short)]= func;

	// Daten in Sendebuffer kopieren
	memcpy(&m_transmit[IV_PACKET_LENGTH_HEAD + m_transmit[3] + sizeof(unsigned char) + sizeof(unsigned short) * 2], buf, count);	
	
	// 10a
	// Daten Verschluesselung
	//if (IsKeyOEM2()) 
	//{
	//	// Packetversion umschalten
	//	m_transmit[IV_PACKET_LENGTH_HEAD + m_transmit[3]] &= ~IV_PACKET_VERSION_4;
	//	m_transmit[IV_PACKET_LENGTH_HEAD + m_transmit[3]] |= IV_PACKET_VERSION_5;

	//	// FixValue setzen
	//	*(unsigned short*)&m_transmit[IV_PACKET_LENGTH_HEAD + m_transmit[3] + sizeof(unsigned char) + sizeof(unsigned short) * 2 + count] = IV_PACKET_DF4_FIXVALUE;
	//	*(unsigned short*)&m_transmit[4] += sizeof(unsigned short);
	//	count += sizeof(unsigned short);
	//
	//	// Daten verschluesseln
	//	DF4(&m_transmit[IV_PACKET_LENGTH_HEAD + m_transmit[3] + sizeof(unsigned char)], sizeof(unsigned short) * 2 + count);
	//}
	// 11.
	// Wird von Transmit-Funktion erledigt

	// Zu Punkt 9: Steuerinformation -> Packezaehler
	// Ist Packet eine Anfrage?
	if (m_transmit[IV_PACKET_LENGTH_HEAD + m_transmit[3]] & IV_PACKET_REQUEST) 
	{
		// Aktuelle Packetnummer
		m_transmit[IV_PACKET_LENGTH_HEAD] = m_packetNumber;
	} 
	else if (m_transmit[IV_PACKET_LENGTH_HEAD + m_transmit[3]] & IV_PACKET_ANSWER) 
	{
		// Packet ist Antwort auf Anfrage!
		// Steuerinformation verfuegbar? 
		if (m_receive[3] > 0) 
		{	
			// Dann in die Antwort uebernehmen.
			m_transmit[IV_PACKET_LENGTH_HEAD] = m_receive[IV_PACKET_LENGTH_HEAD];
		}
	}
	Trace("%s(%i):INFO(1): CChannel4Protocol::TransmitPacket(%i) (%i -> %i, B: %02X, M: %i, F: %i, PacketNummer: %i)", __FILE__, __LINE__, GetID(), GetDeviceID(), GetDeviceIDRemote(),
		m_transmit[IV_PACKET_LENGTH_HEAD + m_transmit[3]], 
		*(unsigned short*)&m_transmit[IV_PACKET_LENGTH_HEAD + m_transmit[3] + sizeof(unsigned char)],
		*(unsigned short*)&m_transmit[IV_PACKET_LENGTH_HEAD + m_transmit[3] + sizeof(unsigned char) + sizeof(unsigned short)],
		m_transmit[IV_PACKET_LENGTH_HEAD]);
#ifdef TRACEPACKET
	int len;
	for (len=0; len < *(unsigned short*)&m_transmit[4]; len++)
	{
		if (len == 0)
		{
			Trace("[%02X ", m_transmit[len]);
		}
		else if (len == (*(unsigned short*)&m_transmit[4] - 1))
		{
			Trace("%02X]", m_transmit[len]);
		}
		else
		{
			Trace("%02X ", m_transmit[len]);
		}
	}
#endif
	Trace("\n");

	// Protokoll-Paket senden
	if ((err = TransmitProtocol(m_transmit, transmitTrials)) != IV_ERROR_SUCCESS)
	{	
		Trace("%s(%i):ERROR(1): CChannel4Protocol::TransmitPacket (%i)\n", __FILE__, __LINE__, err);
		return err;
	}

	// Zeitpunkt des Packetversands halten
	m_timeTransmitPacket = time(NULL);

	// Paketanzahl und Nutzdatengroesse fuer Statistik mitprotokollieren
	m_transmittedPackets++;
	m_transmittedSizeData += count;

	return err;
}

// ##################################################################
/**
 * @brief Lesen eines Packets inklusive Pruefung von Kopf und Gesamt-CRC.
 */
int CChannel4Protocol::ReceiveProtocol(unsigned char *buffer, int receiveTrials)
{
	Trace("%s(%i):INFO(3): CChannel4Protocol::ReceiveProtocol [ 0x%08X, %i ]\n", __FILE__, __LINE__, buffer, receiveTrials);

	int countEscape = 0;
	int received;
	int escape = FALSE, escapeError = FALSE;

	m_clockRx = 0;
	clock_t clockRx = 0;

	int res;
	int error = IV_ERROR_SUCCESS;

	int receiveCount = IV_PACKET_LENGTH_HEAD;
	int receivedCount = 0;

	m_skipCount = 0;

	// Liegt Paketstart breits vor?
	if (m_packetStart) 
	{
		buffer[0] = IV_PACKET_SIGN_START;
		receivedCount++;
		// bereits gelesenen Paketstart uebernommen
		m_packetStart = FALSE;
	}

	// Datenstrom zeichenweise lesen und Paketstartzeichen herausfiltern
	for (; receivedCount < receiveCount; receivedCount++) 
	{
		// Einstieg zum lesen eines Zeichens
		received = 0;

		// Empfangszeit fuer Pakete ermitteln.
		clock_t clockStart = clock();
		if ((res = ReceiveData(&buffer[receivedCount], 1, received, receiveTrials)) != IV_ERROR_SUCCESS) 
		{
			Trace("%s(%i):ERROR(1): CChannel4Protocol::ReceiveProtocol [ ReceiveData fail with: %i ]\n", __FILE__, __LINE__, res);
			return (error != IV_ERROR_SUCCESS) ? error : res;
		}
		clockRx += clock() - clockStart;

		// Empfangene Anzahl Zeichen erhoehen
		m_receivedSize += received;

		// Error abarbeiten
		if (error != IV_ERROR_SUCCESS) 
		{
			// Erstes Zeichen muss Startzeichen sein.
			if (buffer[receivedCount] != IV_PACKET_SIGN_START) 
			{
				// Startzeichen fehlt noch
				Trace("%s(%i):WARNING(1): CChannel4Protocol::ReceiveProtocol [ Skip first Databyte 0x%02X ]\n", __FILE__, __LINE__, buffer[receivedCount]);
				// Maximal zu lesende Zeichen ohne Paketstart
				if (++m_skipCount > IV_SKIPCOUNT) 
				{
					Trace("%s(%i):ERROR(1): CChannel4Protocol::ReceiveProtocol [ Maximal skip count reached: %i ]\n", __FILE__, __LINE__, error);
					return -IV_ERROR_SKIPCOUNT;
				}
				// Emfpang zuruecksetzen
				receivedCount = -1;
				continue;
			} 
			else 
			{
				// Packetstart gelesen
				m_packetStart = TRUE;
			}

			Trace("%s(%i):ERROR(1): CChannel4Protocol::ReceiveProtocol [ %i ]\n", __FILE__, __LINE__, error);
			return error;
		}

		// Erstes Zeichen muss Startzeichen sein.
		if (!receivedCount && buffer[receivedCount] != IV_PACKET_SIGN_START) 
		{	
			// Startzeichen fehlt noch
			Trace("%s(%i):WARNING(1): CChannel4Protocol::ReceiveProtocol [ Skip first Databyte 0x%02X ]\n", __FILE__, __LINE__, buffer[receivedCount]);
			
			// Paketzeit reset
			//ct1 = clock(); 

			// Maximal zu lesende Zeichen ohne Paketstart
			if (++m_skipCount > IV_SKIPCOUNT) 
			{
				Trace("%s(%i):ERROR(1): CChannel4Protocol::ReceiveProtocol [ Maximal skip count reached: %i ]\n", __FILE__, __LINE__, -IV_ERROR_SKIPCOUNT);
				return -IV_ERROR_SKIPCOUNT;
			}

			// Empfang zuruecksetzen
			receivedCount = -1;
			continue;
		}
		else if (receivedCount && buffer[receivedCount] == IV_PACKET_SIGN_START) 
		{
			// Ein Startzeichen wurde im Paket empfangen.
			Trace("%s(%i):ERROR(1): CChannel4Protocol::ReceiveProtocol [ Startbyte in packet: %i ]\n", __FILE__, __LINE__, -IV_ERROR_START_IN_PACKET);
			
			m_packetStart = TRUE;
			return -IV_ERROR_START_IN_PACKET;
		} 
		else if (receivedCount && (buffer[receivedCount] == IV_PACKET_SIGN_ESCAPE)) 
		{
			// Escape bereits eingeschaltet?
			if (escape) 
			{
				Trace("%s(%i):ERROR(1): CChannel4Protocol::ReceiveProtocol [ Double escape: %i ]\n", __FILE__, __LINE__, -IV_ERROR_ESCAPE);

				// Escape ausschalten
				escape = FALSE;
				// Fehler setzen
				error = -IV_ERROR_ESCAPE;
				m_skipCount = 0;
				// Emfpang zuruecksetzen
				receivedCount = -1;
				continue;
			}

			// Escape einschalten
			escape = TRUE;
			// Emfpangszeiger zuruecksetzen
			receivedCount--;
			continue;
		} 
		else if (escape) 
		{
			// Escape ausschalten
			escape = FALSE;

			// Zu escapierendes Zeichen
			switch (buffer[receivedCount]) 
			{
			case 0:
				buffer[receivedCount] = IV_PACKET_SIGN_ESCAPE;
				break;
			case 1:
				buffer[receivedCount] = IV_PACKET_SIGN_START;
				break;
			case 2:
				// Bleibt im Puffer erhalten
				break;
			default:
				Trace("%s(%i):ERROR(1): CChannel4Protocol::ReceiveProtocol [ Unknown escape-sequenz: %i ]\n", __FILE__, __LINE__, -IV_ERROR_ESCAPESEQUENZ);
				escapeError =  -IV_ERROR_ESCAPESEQUENZ;
				m_skipCount = 0;
				// Emfpang zuruecksetzen
				receivedCount = -1;
				continue;
			}

			// Escapezeichen mitzaehlen
			countEscape++;
		}

		// Paketdaten auswerten
		// Es gibt 2 Auswertesituationen.
		// 1. Der Paketkopf wurde komplett empfangen -> Kopf-CRC pruefen. Geamtanzahl Packetdaten setzen.
		// 2. Das gesamte Paket wurde empfangen. -> Gesamt-CRC pruefen
		if ((receivedCount + 1) == receiveCount) 
		{
			// Paketkopf auswerten
			if ((receivedCount + 1) == IV_PACKET_LENGTH_HEAD) 
			{
				// CRC pruefen
				if (GetCRC16(buffer, receivedCount + 1)) 
				{
					Trace("%s(%i):ERROR(1): CChannel4Protocol::ReceiveProtocol [ Bad head-crc: %i ]\n", __FILE__, __LINE__, -IV_ERROR_CRC_HEAD);

					int p = 0; // Fehlerhafte Kopfdaten ausgeben
					for (; p < (receivedCount + 1); p++) 
					{
						// Fehlerhafte Kopfdaten trace
						Trace("%s(%i):WARNING(1): CChannel4Protocol::ReceiveProtocol [ Skip( %3u ) head Databyte 0x%02X ]\n", __FILE__, __LINE__, p+1, buffer[p]);
					}

					error = -IV_ERROR_CRC_HEAD;
					// Emfpang zuruecksetzen
					receivedCount = -1;
					continue;
				}

				// Paketgroesenangabe pruefen
				if (*(unsigned short*)&buffer[4] < IV_PACKET_LENGTH_FRAME ||
					*(unsigned short*)&buffer[4] >= IV_PACKET_LENGTH_MAX) 
				{
					error = -IV_ERROR_PACKETLENGTH;
					// Emfpang zuruecksetzen
					receivedCount = -1;
					continue;
				}

				// Neue Empfangsgroesse setzen
				receiveCount = *(unsigned short*)&buffer[4];
			}
			else // Gesamt-CRC pruefen
			{
				// CRC pruefen
				if (GetCRC16(buffer, receivedCount + 1)) 
				{
					Trace("%s(%i):ERROR(1): CChannel4Protocol::ReceiveProtocol [ Bad packet-crc: %i ]\n", __FILE__, __LINE__, -IV_ERROR_CRC_PACKET);

					int p = 0; // Fehlerhafte Kopf- Paketdaten ausgeben
					for (; p < (receivedCount + 1); p++) 
					{
						// Fehlerhafte Kopfdaten trace
						Trace("%s(%i):WARNING(1): CChannel4Protocol::ReceiveProtocol [ Skip( %3u ) %s Databyte 0x%02X ]\n", __FILE__, __LINE__, p+1, (p < IV_PACKET_LENGTH_HEAD) ? "head" : "paket", buffer[p]);
					}

					error = -IV_ERROR_CRC_PACKET;
					// Emfpang zuruecksetzen
					receivedCount = -1;
					continue;
				}
			}
		} 
	}

	// Gesamtzeit des Lesevorgangs fuer Statistik halten
	m_clockRx = clockRx;
	m_clocksRx += m_clockRx;

	return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Lesen eines Packets inklusive Pruefung von Kopf und Gesamt-CRC.
*/
int CChannel4Protocol::ReceiveTimeboyStation(unsigned char *buffer, int receiveTrials)
{
	Trace("%s(%i):INFO(3): CChannel4Protocol::ReceiveProtocol [ 0x%08X, %i ]\n", __FILE__, __LINE__, buffer, receiveTrials);

	static unsigned int crcLast = 0;
	static int crcRepeated = 0;
	int received;
	int escape = FALSE;

	m_clockRx = 0;
	clock_t clockRx = 0;
	clock_t clockStart = clock();
	clock_t clockEnd = clock() + ((receiveTrials * GetTimeout()) / 1000) * CLOCKS_PER_SEC;

	int res = IV_ERROR_SUCCESS;

	int toReceive = IV_PACKET_LENGTH_FRAME; // Haelt die zu empfangende Anzahl Bytes
	int receivedCount = 0;                  // Haelt die Gesamtanzahl gelesener Bytes in m_tmp
	int headSearches = 0;                   // Haelt die Anzahl der Leseversuche eines Packetanfanges beim Timeboy

	int receivePos = 0;     // Position des Tmp - Lesepuffers
	int packetPos = 0;  // Position des Paketpuffers

	// Enum fuer Kommunikationsstatus.
	enum 
	{
		begin,         // Beginn der Paketauswertung.
		headSearch,    // Suche des Paketbeginns '~'
		headReceive,   // Daten des Paketkopfes werden empfangen.
		headComplete,  // Paketkopf komplett, CRC-Pruefung war OK.
		bodyReceive,   // Paketdaten werden empfangen.
		complete,      // Paket komplett empfangen
		failed
	} receiveStatus; // Aktueller Empfangsstatus

	// Aktuellen Empfangsstatus setzen.
	receiveStatus = begin;

	// Solange Zeichen zu empfangen sind.
	for ( ; receiveStatus != complete && receiveStatus != failed; )
	{

		// Je nach aktuellem Empfangsstatus.
		switch (receiveStatus)
		{
		case begin: // Beginn der Paketauswertung.
			
			// Temp - Lesepuffer zuruecksetzen
			receivePos = 0;
			receivedCount = 0;
			// Auswertungspuffer zuruecksetzen
			packetPos = 0;

			// Zuerst zu lesende Menge auf Maximale Puffergroesse setzen.
			// Die Dockingstation liest an der Folgeadresse das naechste Datenpaket aus dem Timeboy!
			// Wir hier eine kleine Paketgroesse angegeben werden alle Folgepakete mit dieser Laenge geliefert.
			toReceive = MB_SLOT_BUFFER_DATA_SIZE_SMALL;//GetDockingBufferSize(GetDeviceIDRemote());

			// Empfangsstatus wechseln
			receiveStatus = headSearch;
			break;

		case headSearch:
		
			// Erstes Datenbyte uebernehmen
			buffer[0] = m_tmp[0];

			// Erstes Zeichen muss Startzeichen sein.
			if (buffer[0] != IV_PACKET_SIGN_START)
			{	
				Trace("%s(%i):WARNING(1): CChannel4Protocol::ReceiveProtocol [ Paket start wait: 0x%02X ]\n", __FILE__, __LINE__, buffer[0]);

				headSearches++;
				// Abbruch nach mehrfachem Nachlesen.
				if (m_timeboyBaseAddress == 0x70000000 && headSearches > 5)
				{
					// Zu viele Versuche,
					res = -IV_ERROR_SKIPCOUNT;
					// Fehlerzustand annehmen.
					receiveStatus = failed;
					continue;
				}
				else if (m_timeboyBaseAddress == 0x70010000 && headSearches > 30)
				{
					// Zu viele Versuche,
					res = -IV_ERROR_SKIPCOUNT;
					// Fehlerzustand annehmen.
					receiveStatus = failed;
					continue;
				}

				// Zeit fuer Versuche abgelaufen
				//if (clock() > clockEnd)
				//{
				//	// Zu viele Versuche,
				//	res = -IV_ERROR_RECEIVE;
				//	// Fehlerzustand annehmen.
				//	receiveStatus = failed;
				//	continue;
				//}

				// Erneut den Paketstart lesen.
				receiveStatus = begin;
				continue;
			}

			// Paketanfang erfolgt
			receiveStatus = headReceive;
			continue;

		case headReceive:

			packetPos++;
			receivePos++;
			// Naechsten Wert fuer Auswertung uebernehmen
			buffer[packetPos] = m_tmp[receivePos];

			// Paketdaten auswerten
			// Es gibt 2 Auswertesituationen.
			// 1. Der Paketkopf wurde komplett empfangen -> Kopf-CRC pruefen. Geamtanzahl Packetdaten setzen.
			// 2. Das gesamte Paket wurde empfangen. -> Gesamt-CRC pruefen

			// Paketkopf auswerten
			if ((packetPos + 1) == IV_PACKET_LENGTH_HEAD) 
			{
				// CRC pruefen
				if (GetCRC16(buffer, IV_PACKET_LENGTH_HEAD)) 
				{
					Trace("%s(%i):ERROR(1): CChannel4Protocol::ReceiveProtocol [ Bad head-crc: %i ]\n", __FILE__, __LINE__, -IV_ERROR_CRC_HEAD);

					int p = 0; // Fehlerhafte Kopfdaten ausgeben
					for (; p < packetPos; p++) 
					{
						// Fehlerhafte Kopfdaten trace
						Trace("%s(%i):WARNING(1): CChannel4Protocol::ReceiveProtocol [ Skip head checked Databyte 0x%02X ]\n", __FILE__, __LINE__, buffer[p]);
					}

					res = -IV_ERROR_CRC_HEAD;
					// Schnittstelle bereinigen
					receiveStatus = failed;
					continue;
				}

				// Paketgroesenangabe pruefen
				if (*(unsigned short*)&buffer[4] < IV_PACKET_LENGTH_FRAME ||
					*(unsigned short*)&buffer[4] >= IV_PACKET_LENGTH_MAX) 
				{
					res = -IV_ERROR_PACKETLENGTH;
					// Schnittstelle bereinigen
					receiveStatus = failed;
					continue;
				}

				// aktueller Status ist Paketkopf gueltig empfangen.
				receiveStatus = headComplete;
				continue;
			}
			break;

		case headComplete:
			// Anzahl nachzulesender Datenbytes setzen?
			if (*(unsigned short*)&buffer[4] > receivedCount)
			{
				// Neue Empfangsgroesse setzen
				toReceive = *(unsigned short*)&buffer[4];// - receivedCount;
			}

			// Empfangsstatus anpassen
			receiveStatus = bodyReceive;
			break;

		case bodyReceive:

			packetPos++;
			receivePos++;
			// Naechsten Wert fuer Auswertung uebernehmen
			buffer[packetPos] = m_tmp[receivePos];

			if ((packetPos + 1) == *(unsigned short*)&buffer[4]) 
			{		
				// Gesamt - CRC pruefen
				if (GetCRC16(buffer, packetPos + 1)) 
				{
					Trace("%s(%i):ERROR(1): CChannel4Protocol::ReceiveProtocol [ Bad packet-crc: %i ]\n", __FILE__, __LINE__, -IV_ERROR_CRC_PACKET);
					res = -IV_ERROR_CRC_PACKET;
					// Emfpang zuruecksetzen
					receiveStatus = failed;
					continue;
				}

				if (crcLast != GetCRC32(buffer, packetPos + 1))
				{
					// Letzten CRC von Packet merken.
					crcLast = GetCRC32(buffer, packetPos + 1);
					crcRepeated = 0;
				}
				else
				{
					// Wiederholt gleiches Paket gelesen
					if (crcRepeated++ > 3)
					{
						crcLast = 0;
						crcRepeated = 0;
						// Dreimalig gleiches Paket gelesen.
						Trace("%s(%i):ERROR(1): CChannel4Protocol::ReceiveProtocol [ Repeated packet-crc: %i ]\n", __FILE__, __LINE__, -IV_ERROR_CRC_PACKET);
						res = -IV_ERROR_SKIPCOUNT;
						// Emfpang zuruecksetzen
						receiveStatus = failed;
						continue;
					}

				}
				// Paket komplett empfangen
				receiveStatus = complete;
				continue;
			} 
			break;


		default:
			Trace("%s(%i):ERROR(1): CChannel4Protocol::ReceiveProtocol [ receiveStatus unbekannt: %i ]\n", __FILE__, __LINE__, receiveStatus);
		}

		// Sind Zeichen zu lesen?
		if (toReceive > 0)
		{
			// Einstieg zum lesen eines Teilpaketes
			received = 0;

			//if ((res = ReadBytes(GetDeviceIDRemote(), GetTimeboyStationBaseAddress() + receivedCount, &m_tmp[receivedCount], toReceive)) != IV_ERROR_SUCCESS)
			if ((res = m_timeboyStation.ReadBytes(GetDeviceIDRemote(), m_timeboyBaseAddress, receivedCount, m_tmp, toReceive)) != IV_ERROR_SUCCESS)
			{
				Trace("%s(%i):ERROR(1): CChannel4Protocol::ReceiveProtocol [ ReadBytes fail with: %i ]\n", __FILE__, __LINE__, res);
				return res;
			}

			// Empfangsanzahl angleichen
			received = toReceive;

			// Gesamtanzahl gelesener Bytes halten
			receivedCount += received;

			// Anzahl zu lesender Bytes wurde erreicht, somit ist diese nun 0
			toReceive -= received;

			// Statistik: Empfangene Anzahl Zeichen erhoehen
			m_receivedSize += received;
		}
	}

	// Gesamtzeit des Lesevorgangs fuer Statistik halten
	m_clockRx = clock() - clockStart;
	m_clocksRx += m_clockRx;

	return res;
}

// ##################################################################
/**
 * @brief Lesen eines Packets aus der Kommunikationsschnittstelle.
 *
 * @param deviceIDRemote Anzusprechende Geraetenummer.
 */
int CChannel4Protocol::ReceivePacket(short deviceID, unsigned char &type, short &module, short &func, unsigned char *buf, int &count, int receiveTrials)
// Empfangen und verarbeiten eines Pakets
//  deviceIDRemote -> Adresse (Geraetenummer) des Empfaengers
//  type     -> Pakettyp (Anfrage/Antwort +? Error)
//  module   -> Nummer des anzusprechenden Modules
//  func -> Nummer der Funktion im Module
//	buf	     -> Buffer fuer Funktionsparameter.
//	count    -> Anzahl der Funktionsparameter in Byte. Rueckgabewert bei Parameterempfang.
//
//  return -> <0 -> Kommunikationsfehler
//             0 -> senden und empfangen und ausfuehren erfolgreich
//			  >0 -> module-/Funktionsfehler
{
	Trace("%s(%i):INFO(3): CChannel4Protocol::ReceivePacket [ %i, %i, %i, 0x%08X, %i ]\n", __FILE__, __LINE__, deviceID, module, func, buf, count);

	int err = IV_ERROR_SUCCESS;

	// Adresse setzen
	SetDeviceID(deviceID);

/* 
1.	(  8 Bit): Startzeichen ("~", 0x7E)
2.	(  8 Bit): Empfngeradresse
3.	(  8 Bit): Absenderadresse
4.	(  8 Bit): Steuerinformationslnge
5.	( 16 Bit): Paketgesamtlnge
6.	(  8 Bit): Flags - fr Verschlsselung und Flashprfung (z. B. fr aktive Verschlsselung)
7.	(  8 Bit): Reserve
8.	( 16 Bit): CRC ber die ersten 7 Felder
9.	(n*8 Bit): Steuerinformationsdaten (in Bezug auf Feld 5)
10.	(n*8 Bit): Zu transportierende Daten
11.	( 16 Bit): CRC ber Felder 1 - 10.
*/

receive_again:
	// Wird ueber Dockingstation kommuniziert?
	if (IsTimeboySlot(GetDeviceIDRemote()))
	{
		// Protokoll-Paket empfangen
		if ((err = ReceiveTimeboyStation(m_receive, receiveTrials)) != IV_ERROR_SUCCESS) 
		{
			Trace("%s(%i):ERROR(1): CChannel4Protocol::ReceivePacket (%i)\n", __FILE__, __LINE__, err);
			return err;
		}
	}
	else
	{
		// Protokoll-Paket empfangen
		if ((err = ReceiveProtocol(m_receive, receiveTrials)) != IV_ERROR_SUCCESS) 
		{
			Trace("%s(%i):ERROR(1): CChannel4Protocol::ReceivePacket (%i)\n", __FILE__, __LINE__, err);
			return err;
		}
	}

	Trace("%s(%i):INFO(1): CChannel4Protocol::ReceivePacket(%i) (%i <- %i, B: %02X, M: %i, F: %i, PacketNummer: %i)", __FILE__, __LINE__, GetID(), GetDeviceID(), m_receive[2],
		m_receive[IV_PACKET_LENGTH_HEAD + m_receive[3]], 
		*(unsigned short*)&m_receive[IV_PACKET_LENGTH_HEAD + m_receive[3] + sizeof(unsigned char)],
		*(unsigned short*)&m_receive[IV_PACKET_LENGTH_HEAD + m_receive[3] + sizeof(unsigned char) + sizeof(unsigned short)],
		(m_receive[3]) ? m_receive[IV_PACKET_LENGTH_HEAD] : 0);
#ifdef TRACEPACKET
	int len;
	for (len=0; len < *(unsigned short*)&m_receive[4]; len++)
	{
		if (len == 0)
		{
			Trace("[%02X ", m_receive[len]);
		}
		else if (len == (*(unsigned short*)&m_receive[4] - 1))
		{
			Trace("%02X]", m_receive[len]);
		}
		else
		{
			Trace("%02X ", m_receive[len]);
		}
	}
#endif
	Trace("\n");

	//	Empfaengeradresse pruefen
	if (IV_PACKET_DEVICEID(m_receive) != GetDeviceID()) 
	{
		err = -IV_ERROR_ADDRESS;
		Trace("%s(%i):WARNING(1): CChannel4Protocol::ReceivePacket (Wrong DeviceID %i for %i)\n", __FILE__, __LINE__, IV_PACKET_DEVICEID(m_receive), GetDeviceID());
		// Handelt es sich um keine Docking,
		// weil dort kein Datenstromabriss erfolgt, wuerde eine Entlosschleife entstehen.
		if (IsTimeboySlot(GetDeviceIDRemote()))
		{
			return err;
		}

		goto receive_again;
	}

	// Variable fr die Flags in Member halten
	m_lastFlags = IV_PACKET_FLAGS(m_receive);
	/////////////////////////////////
	// Die Flags der Flashprfen prfen
	if (m_lastFlags & IV_PACKET_FLAG_FLASH_SERVICE)
	{	// Flash Service: Globaler Wear-Level ist >= 6
		Trace("%s(%i):WARNING(1): CChannel4Protocol::ReceivePacket (Flash Service Flag is set 0x%02X)\n", __FILE__, __LINE__, m_lastFlags);
		// Flash Service occurce
	}
	if (m_lastFlags & IV_PACKET_FLAG_FLASH_WARNING)
	{	// Flash Warning: CRC Fehler im Ringspeicher oder Listenbereich. (Setup und Listen neu bertragen)
		Trace("%s(%i):WARNING(1): CChannel4Protocol::ReceivePacket (Flash Warning Flag is set 0x%02X)\n", __FILE__, __LINE__, m_lastFlags);
		// Flash Warning occurce
	}
	if (m_lastFlags & IV_PACKET_FLAG_FLASH_LOCKED)
	{	// Flash Error: Keine freien Seiten mehr verfgbar, Gert befindet sich im Bedienstopp.
		Trace("%s(%i):WARNING(1): CChannel4Protocol::ReceivePacket (Flash Error Flag is set 0x%02X)\n", __FILE__, __LINE__, m_lastFlags);
		// Flash Error occurce
	}

	// Zu Punkt 9: Steuerinformation Packetzaehler
	// Ist Packet eine Antwort auf Anfrage?
	if (m_receive[IV_PACKET_LENGTH_HEAD + m_receive[3]] & IV_PACKET_ANSWER) 
	{
		// Steuerinformation verfuegbar?
		if (m_receive[3] > 0) 
		{
			// Ist Packetnummer der letzten Anfrage?
			if (m_receive[IV_PACKET_LENGTH_HEAD] == m_packetNumber) 
			{
				// Packetnummer erhoehen
				m_packetNumber++;
			} 
			else 
			{
				err = -IV_ERROR_ADDRESS;
				Trace("%s(%i):WARNING(1): CChannel4Protocol::ReceivePacket (Skip wrong packet number [%i instead %i])\n", __FILE__, __LINE__, m_receive[IV_PACKET_LENGTH_HEAD], m_packetNumber);
				// Handelt es sich um keine Docking,
				// weil dort kein Datenstromabriss erfolgt, wuerde eine Entlosschleife entstehen.
				if (IsTimeboySlot(GetDeviceIDRemote()))
				{
					return err;
				}
				goto receive_again;
			}
		}
	}

	// Packettyp uebernehmen
	type = m_receive[IV_PACKET_LENGTH_HEAD + m_receive[3]];
	// Anzahl der Datenbyte
	count = *(unsigned short*)&m_receive[4] - IV_PACKET_LENGTH_FRAME - m_receive[3] - sizeof(unsigned char) - sizeof(unsigned short) * 2;

	// Packetdaten ggf. entschluesseln
	//if ((m_receive[6] & IV_PACKET_FLAG_DF4_KEY_OEM2)) 
	//{
	//	DF4(&m_receive[IV_PACKET_LENGTH_HEAD + m_receive[3] + sizeof(unsigned char)], 2 * sizeof(unsigned short) + count);

	//	if (*(unsigned short*)&m_receive[IV_PACKET_LENGTH_HEAD + m_receive[3] + sizeof(unsigned char) + sizeof(unsigned short) * 2 + (count - sizeof(unsigned short))] != IV_PACKET_DF4_FIXVALUE) 
	//	{
	//		// Entschluesselter Wert stimmt nicht
	//		Trace("%s(%i):ERROR(1): CChannel4Protocol::ReceivePacket\n", __FILE__, __LINE__);
	//	}

	//	count -= sizeof(unsigned short);
	//}

	// Modul und Funktion uebernehmen
	module = *(unsigned short*)&m_receive[IV_PACKET_LENGTH_HEAD + m_receive[3] + sizeof(unsigned char)];
	func = *(unsigned short*)&m_receive[IV_PACKET_LENGTH_HEAD + m_receive[3] + sizeof(unsigned char) + sizeof(unsigned short)];

	// Daten uebernehmen
	memcpy(buf, &m_receive[IV_PACKET_LENGTH_HEAD + m_receive[3] + sizeof(unsigned char) + sizeof(unsigned short) * 2], count);

	// Zeitpunkt des Packetempfangs halten
	m_timeReceivePacket = time(NULL);

	// Paketanzahl und Nutzdatengroesse fuer Statistik mitprotokollieren
	m_receivedPackets++;
	m_receivedSizeData += count;

	// Call-Back Funktion aufrufen, ggf. abbruch.
	if (m_callBack1) 
	{
		if (!m_callBack1()) 
		{	
			Trace("%s(%i):ERROR(1): CChannel4Protocol::ReceivePacket [ CallBack abbruch ]\n", __FILE__, __LINE__);
			return -IV_ERROR_BREAK;
		}
	}

	return err;
}

// ##################################################################
/**
 * @brief Remoteaufruf einer lokalen Modulfunktion.
 */
int CChannel4Protocol::Call(short module, short func, unsigned char *data, int &count)
{
	// Fehler: Das geforderte Modul wird nicht unterstuetzt.
	return 0x0100;
}

// ##################################################################
/**
 * @brief Ausfuehrung einer Modulfunktion in ihrem jeweiligen Kontext.
 *
 */
int CChannel4Protocol::Execute(short context, short module, short func, unsigned char *buf, int &count, int contextValue)
{
	Trace("%s(%i):INFO(3): CChannel4Protocol::Execute [ %i, %i, %i, 0x%08X, %i, %i ]\n", __FILE__, __LINE__, context, module, func, buf, count, contextValue);

	int trials, receiveTrials, retries;
	int txCount = 0, rxCount;
	unsigned char txTmp[IV_PACKET_LENGTH_DATA], rxTmp[IV_PACKET_LENGTH_DATA];
	short rxModule, rxFunction;
	unsigned char type = 0;
	int err = IV_ERROR_SUCCESS;

	switch(context) {

	case IV_CONTEXT_TRANSMIT_ANSWER: // Einfacher Versant eines Packetes
	case IV_CONTEXT_TRANSMIT_ANSWER_ERROR: // Einfacher Versant eines Packetes
		// Temppuffer fuer Transmit-Packet mit Daten fuellen.
		txCount = count;
		memcpy(txTmp, buf, txCount);
		retries = (contextValue == -1) ? IV_TRANSMIT_TRIALS : contextValue;
		receiveTrials = 0;
		m_timeboyBaseAddress = 0x70010000;
		break;

	case IV_CONTEXT_MASTER: // Ausfhrung (Versant und Empfang) eines Befehls
		// Temppuffer fuer Transmit-Packet, falls ein Request des Geraetes dazwischen kommt und
		// die selben Daten nochmals gesendet werden.
		txCount = count;
		memcpy(txTmp, buf, txCount);
		retries = (contextValue == -1) ? IV_TRANSMIT_TRIALS : contextValue;
		receiveTrials = IV_RECEIVE_TRIALS;
		m_timeboyBaseAddress = 0x70000000;
		break;

	case IV_CONTEXT_SLAVE: // Empfang und Ausfhrung eines Befehls
		retries = (contextValue == -1) ? IV_RECEIVE_TRIALS : contextValue; 
		receiveTrials = 1;
		m_timeboyBaseAddress = 0x70010000;
		break;

	default:
		Trace("%s(%i):ERROR(1): CChannel4Protocol::Execute [ Unbekannter Context ]\n", __FILE__, __LINE__);
		return -IV_ERROR_FAIL;
	}

	// Einen Versuch sicherstellen
	if (retries <= 0) {	retries = 1; }

	for (trials=0; trials < retries; trials++) 
	{
		// Ist ein Kommunikationsbefehl abzuarbeiten?
		if (context == IV_CONTEXT_MASTER || context == IV_CONTEXT_TRANSMIT_ANSWER || context == IV_CONTEXT_TRANSMIT_ANSWER_ERROR) 
		{
			unsigned char type = (context == IV_CONTEXT_MASTER) ? IV_PACKET_REQUEST : IV_PACKET_ANSWER;
			if (context == IV_CONTEXT_TRANSMIT_ANSWER_ERROR) type |= IV_PACKET_ERROR;
			// Befehlspacket senden 
			if ((err = TransmitPacket(GetDeviceIDRemote(), type, module, func, txTmp, txCount)) != IV_ERROR_SUCCESS) 
			{
				Trace("%s(%i):ERROR(1): CChannel4Protocol::Execute [ TransmitPacket (%i, %i) ]\n", __FILE__, __LINE__, context, err);
				break;
			}

			// Kein lesen erforderlich?
			if (context != IV_CONTEXT_MASTER) break;
		} 

receive_again:
		// Befehlspacket empfangen
		if ((err = ReceivePacket(GetDeviceID(), type, rxModule, rxFunction, rxTmp, rxCount, receiveTrials)) != IV_ERROR_SUCCESS) 
		{
			Trace("%s(%i):ERROR(1): CChannel4Protocol::Execute [ ReceivePacket (%i) ]\n", __FILE__, __LINE__, err);
			if (err == -IV_ERROR_BREAK) { break; } // Abbruch durch Call-Back (Anwender)
			if (context == IV_CONTEXT_MASTER)
			{
				// Anzahl unbeantworteter Pakete fuer Statistik halten
				m_receivedPacketsReject++;
			}
			else if (context == IV_CONTEXT_SLAVE)
			{
				switch (err)
				{
				case IV_ERROR_CRC_PACKET: // Gesamt-CRC des Packets ist falsch.
					break;
				case IV_ERROR_PACKETLENGTH: // Gesamt-Laenge des Packets ist falsch.
					break;
				}
			}
			if (err == -IV_ERROR_CHANNEL_OPEN_REMOTE) 
			{ // Remote hat die Verbindung geschlossen
				break; 
			} 
			continue;
		}

		// Packetbefehl auswerten
		switch (type) 
		{
		case (IV_PACKET_VERSION_4|IV_PACKET_REQUEST):
//		case (IV_PACKET_VERSION_5|IV_PACKET_REQUEST): // Verschluesselt
			{
			
			short deviceIDRemote = SetDeviceIDRemote(m_receive[2]);

			// BUGFIX
			// Der Bootloader liefert als aktive Antwort auf IV_FUNCTION_BEGIN_SET_SEGMENT anstatt
			// IV_FUNCTION_BEGIN_SET_SEGMENT_OK ebenfalls den Wert IV_FUNCTION_BEGIN_SET_SEGMENT
			// Ist aktuell Warten auf diese Aktive Antwort?
			if (context == IV_CONTEXT_SLAVE && module == 0x0001/*IV_MODULE_SYSTEM*/ && func == 0x0010/*IV_FUNCTION_BEGIN_SET_SEGMENT_OK*/) 
			{
				// Korrektur des Funktionswertes, wenn ntig
				rxFunction = func;
			}

			// Funktion der entsprechenden Modulklasse aufrufen.
			// Die Routinen senden die Antwort selber, nur Fehler werden nachfolgend fuer
			// alle gemeinsam abgehandelt.
			if ((err = Call(rxModule, rxFunction, rxTmp, rxCount)) != IV_ERROR_SUCCESS) 
			{	
				Trace("%s(%i):ERROR(1): CChannel4Protocol::Execute [ Call(%i, %i, 0x%08X, %i) ]\n", __FILE__, __LINE__, rxModule, rxFunction, rxTmp, rxCount);

				// Fehler als Antwort senden.
				*(short*)rxTmp = (short)err;
				if ((err = TransmitPacket(GetDeviceIDRemote(), IV_PACKET_ANSWER | IV_PACKET_ERROR, rxModule, rxFunction, rxTmp, sizeof(short))) != IV_ERROR_SUCCESS) 
				{	
					Trace("%s(%i):ERROR(1): CChannel4Protocol::Execute [ TransmitPacket (%i) ]\n", __FILE__, __LINE__, err);
					SetDeviceIDRemote(deviceIDRemote);
					return err;
				}
			}

			// Auf Vorgaenger wechseln.
			SetDeviceIDRemote(deviceIDRemote);
			
			// Aktive Antwort OK?
			if ((context == IV_CONTEXT_SLAVE) && (module != IV_MODULE_BASE && rxModule == module && rxFunction == func)) 
			{
				// Datenkopie uebernehmen
				memcpy(buf, rxTmp, rxCount);
				count = rxCount;
				err = IV_ERROR_SUCCESS;
			} 
			else if ((context == IV_CONTEXT_MASTER) || 
				(context == IV_CONTEXT_SLAVE && module != IV_MODULE_BASE)) 
			{
				// Wurde Befehl gesendet und die Antwort ist noch zu lesen,
				// oder wird noch auf die richtige aktive Antwort gewartet.
				Trace("%s(%i):INFO(1): CChannel4Protocol::Execute [ Receive again ]\n", __FILE__, __LINE__);
				goto receive_again;
			}
			break;
			}
			
		case (IV_PACKET_VERSION_4|IV_PACKET_REQUEST|IV_PACKET_ERROR):
		case (IV_PACKET_VERSION_5|IV_PACKET_REQUEST|IV_PACKET_ERROR): 
			{
					
			if (rxCount != sizeof(short)) 
			{
				// Falsche Parameteranzahl
				Trace("%s(%i):ERROR(1): CChannel4Protocol::Execute [ Fehlerrequest (falsche Parameteranzahl) ]\n", __FILE__, __LINE__);
			}

			// Fehlercode zuweisen
			short e = *(short*)rxTmp;

			// Fehler beantworten.
			if ((err = TransmitPacket(m_receive[2], IV_PACKET_ANSWER | IV_PACKET_ERROR, rxModule, rxFunction, m_data, sizeof(short))) != IV_ERROR_SUCCESS) 
			{	
				Trace("%s(%i):ERROR(1): CChannel4Protocol::Execute [ TransmitPacket (%i) ]\n", __FILE__, __LINE__, err);
				return err;
			}

			// Aktive Antwort OK?
			if ((context == IV_CONTEXT_SLAVE) && (module != IV_MODULE_BASE && rxModule == module && rxFunction == func)) 
			{
				// Nur als Aktive Antwort sinnvoll!!!
				return e;
			} 
			else 
			{
				// Unerwartete Anfrage erhalten
				Trace("%s(%i):ERROR(1): CChannel4Protocol::Execute [ Fehlerrequest (Unerwartete Anfrage) ]\n", __FILE__, __LINE__);
			}
			break;
			}

		case (IV_PACKET_VERSION_4|IV_PACKET_ANSWER):
		case (IV_PACKET_VERSION_5|IV_PACKET_ANSWER):
			
			// Antwort auf zuvor gesendeten Befehl?
			if (context == IV_CONTEXT_MASTER) 
			{
				// Antwort von gefordertem Absender?
				if (IV_PACKET_DEVICEID_REMOTE(m_receive) != GetDeviceIDRemote()) 
				{
					Trace("%s(%i):ERROR(1): CChannel4Protocol::Execute [ Falscher Absender ]\n", __FILE__, __LINE__);
					return -IV_ERROR_ADDRESS;
				}

				// Antwort auf aktuelles Modul/Function
				if (rxModule != module || rxFunction != func) 
				{
					Trace("%s(%i):ERROR(1): CChannel4Protocol::Execute [ Falsche Module/Function ]\n", __FILE__, __LINE__);
					goto receive_again;
				}
				
				// Daten in Buffer der Applikation kopieren
				memcpy(buf, rxTmp, rxCount);
				count = rxCount;
				err = IV_ERROR_SUCCESS;
			} 
			else 
			{
				// Unerwartete Antwort erhalten
				Trace("%s(%i):ERROR(1): CChannel4Protocol::Execute [ Unerwartete Antwort ]\n", __FILE__, __LINE__);
			}
			break;
			
		case (IV_PACKET_VERSION_4|IV_PACKET_ERROR):
		case (IV_PACKET_VERSION_4|IV_PACKET_ANSWER|IV_PACKET_ERROR):
		case (IV_PACKET_VERSION_5|IV_PACKET_ANSWER|IV_PACKET_ERROR):

			// Antwort auf zuvor gesendeten Befehl?
			if (context == IV_CONTEXT_MASTER) 
			{
				// Antwort von gefordertem Absender?
				if (IV_PACKET_DEVICEID_REMOTE(m_receive) != GetDeviceIDRemote()) 
				{
					Trace("%s(%i):ERROR(1): CChannel4Protocol::Execute [ Falscher Absender ]\n", __FILE__, __LINE__);
					return -IV_ERROR_ADDRESS;
				}

				// Antwort auf aktuelles Modul/Function
				if (rxModule != module || rxFunction != func) 
				{
					Trace("%s(%i):ERROR(1): CChannel4Protocol::Execute [ Falsche Module/Function ]\n", __FILE__, __LINE__);
					goto receive_again;
				}

				// Fehlercode uebernehmen
				err = *(short*)rxTmp;

				// Pruefen ob Modulfehler vorliegt und die Anfrage erneut gesendet werden muss.
				// PROT_MODULE_ERROR_CRC                    (0x0700) // Gesamt CRC von Packet ist falsch
				// PROT_MODULE_ERROR_CONTROL                (0x0800) // Zu wenig Steuerinformationen
				// PROT_MODULE_ERROR_PROTOCOL               (0x0900) // Protokollversion passt nicht
				// PROT_MODULE_ERROR_NO_QUERY               (0x0A00) // Anfrage erwartet, Antwort erhalten
				// PROT_MODULE_ERROR_BUSY                   (0x0B00) // Antwort erwartet, Anfrage erhalten
				// PROT_MODULE_ERROR_SIZE                   (0x0C00) // Packetgroesse ueberschreitet reservierte Puffergroesse
				if (err == 0x0700 || err == 0x0B00) 
				{
					Trace("%s(%i):ERROR(1): CChannel4Protocol::Execute [ Modulfehler %04X, wiederholtes senden. ]\n", __FILE__, __LINE__, err);
					trials--; // Nochmaliges senden des Packets
					continue;
				}

			} 
			else 
			{
				// Unerwartete Antwort erhalten
				Trace("%s(%i):ERROR(1): CChannel4Protocol::Execute [ Unerwartete Antwort (context: %i, modul: %i, func: %i) ]\n", __FILE__, __LINE__, context, module, func);
			}
			break;
			
		default:
			Trace("%s(%i):ERROR(1): CChannel4Protocol::Execute [ Unbekannter Packettyp: %02X ]\n", __FILE__, __LINE__, type);
			goto receive_again;
		}

		// Schleife nach erfolgreicher Abarbeitung verlassen.
		break;
	}

	if (context == IV_CONTEXT_MASTER) 
	{
		double tm = (double)((double)m_clockTx + (double)m_clockRx) / (double)CLOCKS_PER_SEC;
		//double clockTx = (double)m_clockTx / (double)CLOCKS_PER_SEC;
		//double clockRx = (double)m_clockRx / (double)CLOCKS_PER_SEC;
		m_qs = (double)((double)m_clocksTx + (double)m_clocksRx) / (double)CLOCKS_PER_SEC;;
		
		if (!(m_receivedPackets % 10)) 
		{	
			double saveVal1 = (m_transmittedSize) ? ((double)(m_transmittedSize - m_transmittedSizeData) / (double)m_transmittedSize) * 100 : 0;
			double saveVal2 = (m_receivedSize) ? ((double)(m_receivedSize - m_receivedSizeData) / (double)m_receivedSize) * 100 : 0;
			double saveVal3 = ((m_transmittedSize + m_receivedSize) == 0) ? 0 : double((m_transmittedSize - m_transmittedSizeData) + (m_receivedSize - m_receivedSizeData)) / (double)(m_transmittedSize + m_receivedSize) * 100;
			double saveVal4 = (m_receivedPackets == 0) ? 0 : (double)m_qs / (double)m_receivedPackets;

			//Trace("%s(%i):INFO(1): CChannel4Protocol::ReceiveProtocol [ QS txBuf(%i ; %i / %i) [sendezeit+empfangszeit: %f][ empfangszeit: %f]{gesamt(empfangszeit/packetanzahl): %f} ]\n", __FILE__, __LINE__, (receivedCount+masked), (double)((double)ct2 - (double)m_ct) / (double)CLOCKS_PER_SEC, tm, (double)qs / (double)count);
			Trace("%s(%i):INFO(4): CChannel4Protocol::ReceiveProtocol [ txBuf: %i; %i/%i; %.2f; rxBuf: %i, %i; %i/%i; %.2f; s(txBuf/rxBuf): %i/%i; %.2f; t(txBuf+rxBuf): %f; t(rxBuf): %f; s(t(rxBuf+txBuf)/pa): %f; ]\n", __FILE__, __LINE__, 
				m_transmittedPackets, m_transmittedSize, m_transmittedSizeData, saveVal1, 
				m_receivedPackets, m_receivedPacketsReject, m_receivedSize, m_receivedSizeData, saveVal2, 
				m_transmittedSize + m_receivedSize, m_transmittedSizeData + m_receivedSizeData, saveVal3,
				tm, (double)m_clockRx / (double)CLOCKS_PER_SEC, saveVal4);
/*
			double val1 = (m_transmittedSize == 0) ? 0 : ((double)(m_transmittedSize - m_transmittedSizeData) / (double)m_transmittedSize) * 100;
			double val2 = (m_receivedSize == 0) ? 0 : ((double)(m_receivedSize - m_receivedSizeData) / (double)m_receivedSize) * 100;
			double val3 = ((m_transmittedSize + m_receivedSize) == 0) ? 0 : double((m_transmittedSize - m_transmittedSizeData) + (m_receivedSize - m_receivedSizeData)) / (double)(m_transmittedSize + m_receivedSize) * 100;
			double val4 = (m_receivedPackets == 0) ? 0 : (double)m_qs / (double)m_receivedPackets;

			Trace("%s(%i):INFO(4): CChannel4Protocol::ReceiveProtocol [ transmit { time %f, packetcount %i; size %i / datasize %i; overhead in prozent %.2f%% } ]\n", __FILE__, __LINE__, clockTx, m_transmittedPackets, m_transmittedSize, m_transmittedSizeData, val1);
			Trace("%s(%i):INFO(4): CChannel4Protocol::ReceiveProtocol [ receive { time %f, packetcount %i, packetcount unbeantwortet %i; size %i / datasize %i; overhead %.2f%% } ]\n", __FILE__, __LINE__, clockRx, m_receivedPackets, m_receivedPacketsReject, m_receivedSize, m_receivedSizeData, val2);
			Trace("%s(%i):INFO(4): CChannel4Protocol::ReceiveProtocol [ summe(transmit + receive) { size %i (%.3fkByte) / datasize %i (%.3fkByte); overhead %.2f%% } ]\n", __FILE__, __LINE__, m_transmittedSize + m_receivedSize, (double)(m_transmittedSize + m_receivedSize) / (double)1024, m_transmittedSizeData + m_receivedSizeData, (double)(m_transmittedSizeData + m_receivedSizeData) / (double)1024, val3);
			Trace("%s(%i):INFO(4): CChannel4Protocol::ReceiveProtocol [ time summe(transmit + receive) { %f; durchschnittszeit pro einzelbefehl: %f } ]\n", __FILE__, __LINE__, tm, val4);
*/
		}
	}

	// Wurde die maximale Anzahl moeglicher Versuche erreicht?
	if (trials == retries) 
	{
		Trace("%s(%i):ERROR(1): CChannel4Protocol::Execute [ Maximalanzahl der Versuche erreicht. ]\n", __FILE__, __LINE__);
		
		// Packetzaehler erhoehen!
		m_packetNumber++;

//		return -IV_ERROR_NO_REQUEST; // Die aktive Antwort kam nicht nach der vorgegebenen Dauer
	}

	return err;
}


//#########################################################
//####################### ERROR ###########################
//#########################################################

// ##################################################################
/**
 * @brief Die gesamte Fehlerbehandlung zuruecksetzen.
 */
void CChannel4Protocol::ResetError()
{
	Trace("%s(%i):INFO(3): CChannel4Protocol::ResetError\n", __FILE__, __LINE__);

	// Fehlernummern loeschen
	m_errorCommunicationNum = m_errorModuleNum = m_errorFunctionNum = 0;
	// Puffer die in Gebrauch waren, loeschen.
	if (m_errorCommunication[0])
	{
		memset(m_errorCommunication, 0, sizeof(m_errorCommunication));
	}
	if (m_errorModule[0])
	{
		memset(m_errorModule, 0, sizeof(m_errorModule));
	}
	if (m_errorFunction[0])
	{
		memset(m_errorFunction, 0, sizeof(m_errorFunction));
	}
}

// ##################################################################
/**
 * @brief Logen der textuellen Fehlerentsprechung.
 *
 */
void CChannel4Protocol::TraceError(short deviceIDRemote, short module, short func)
{
	Trace("%s(%i):INFO(3): CChannel4Protocol::TraceError [ %i, %i, %i ]\n", __FILE__, __LINE__, deviceIDRemote, module, func);

	if (m_result < 0)
	{		
		// Fehlernummern setzen
		m_errorCommunicationNum = -m_result;
		// Fehlertext ermitteln und eintragen.
		GetErrorCommunication(m_errorCommunicationNum, m_errorCommunication);

		Trace("%s(%i):INFO(2): CChannel4Protocol::TraceError ( %04X )[%s]\n", __FILE__, __LINE__, -m_result & 0xFFFF, m_errorCommunication);
	}
	else if (m_result > 0)
	{
		// Fehlertext ermitteln und eintragen.
		if (IV_RESULT_MODULE(m_result))
		{
			// Fehlernummern setzen
			m_errorModuleNum = IV_RESULT_MODULE(m_result);
			// Modulefehlertext ermitteln
			GetErrorModule(module, m_errorModuleNum, m_errorModule);

			Trace("%s(%i):INFO(2): CChannel4Protocol::TraceError ( %04X )[%s]\n", __FILE__, __LINE__, m_result & 0xFFFF, m_errorModule);
		}
		else
		{
			// Fehlernummern setzen
			m_errorFunctionNum = IV_RESULT_FUNCTION(m_result);
			// Funktionsfehlertext ermitteln
			GetErrorFunction(module, func, m_errorFunctionNum, m_errorFunction);

			Trace("%s(%i):INFO(2): CChannel4Protocol::TraceError ( %04X )[%s]\n", __FILE__, __LINE__, m_result & 0xFFFF, m_errorFunction);
		}

	}
}

// ##################################################################
/**
 * @brief Ermittlung des Fehlertextes zum Fehlercode fuer Kommunikationsfehler.
 */
void CChannel4Protocol::GetErrorCommunication(short error, char *text)
{
	Trace("%s(%i):INFO(3): CChannel4Protocol::GetErrorCommunication [ %i, 0x%08X ]\n", __FILE__, __LINE__, error, text);

	switch ( error )
	{
	case IV_ERROR_SUCCESS: sprintf(text, "( %i ): Kein Fehler", error); break;
	case IV_ERROR_CHANNEL_OPEN: sprintf(text, "( %i ): Schnittstelle geschlossen oder konnte nicht geffnet werden.", error); break;
	case IV_ERROR_UNREACHABLE: sprintf(text, "( %i ): Gert antwortet nicht auf Anfrage.", error); break;
	case IV_ERROR_TRANSMIT: sprintf(text, "( %i ): Es konnten keine Daten in die Schnittstelle geschrieben werden.", error); break;
	//case IV_ERROR_RECEIVE: sprintf(text, "( %i ): Es konnten keine Daten von der Schnittstelle gelesen werden.", error); break;
	case IV_ERROR_FILE_OPEN: sprintf(text, "( %i ): Datei konnte nicht geffnet werden.", error); break;
	case IV_ERROR_CRC_PACKET: sprintf(text, "( %i ): Prfsumme (CRC) von Protokollpaket stimmt nicht.", error); break;
	case IV_ERROR_ADDRESS: sprintf(text, "( %i ): Die Adressen (Sender/Empfnger) des Protokollpaketes sind fehlerhaft.", error); break;
	case IV_ERROR_COMMAND: sprintf(text, "( %i ): Protokollbefehl der vierten Schicht ist falsch oder unbekannt.", error); break;
	//case IV_ERROR_PACKETLENGTH: sprintf(text, "( %i ): Die angegebene Paketlaenge stimmt nicht ueberein.", error); break;
	case IV_ERROR_BREAK: sprintf(text, "( %i ): Die Kommunikation wurde durch den Anwender (CallBack-Funktion) abgebrochen.", error); break;
	case IV_ERROR_FUNCTION_UNAVAILABLE: sprintf(text, "( %i ): Funktion steht im vorliegenden Fall nicht zur Verfuegung.", error); break;
	default: sprintf(text, "( %i ): Nicht eingetragener Fehler!", error); break;
	}
}

// ##################################################################
/**
 * @brief Ermittlung des Fehlertextes zum Fehlercode fuer Modulfehler.
 */
void CChannel4Protocol::GetErrorModule(short module, short error, char *text)
{
	Trace("%s(%i):INFO(3): CChannel4Protocol::GetErrorModule [ %i, %i, 0x%08X ]\n", __FILE__, __LINE__, module, error, text);

	// Pruefen ob Modulfehler vorliegt und die Anfrage erneut gesendet werden muss.
	switch ( error )
	{
	case 1: sprintf(text, "( %i ): Das geforderte Modul wird nicht untersttzt.", error); break;
	case 2: sprintf(text, "( %i ): Die geforderte Funktion wird nicht untersttzt.", error); break;
	case 3: sprintf(text, "( %i ): Es wurden bei der Funktion zu wenige Paramteter angegeben.", error); break;
	case 4: sprintf(text, "( %i ): Es wurden bei der Funktion zu viele Paramteter angegeben.", error); break;
//	case 5: sprintf(text, "( %i ): .", error); break;
//	case 6: sprintf(text, "( %i ): .", error); break;
	case 7: sprintf(text, "( %i ): Gesamt CRC des Packetes ist falsch..", error); break;
	case 8: sprintf(text, "( %i ): Es werden zustzliche Steuerinformationen bentigt.", error); break;
	case 9: sprintf(text, "( %i ): Die gesendete Protokollversion ist unbekannt.", error); break;
	case 10: sprintf(text, "( %i ): Es wurde eine Anfrage erwartet.", error); break;
	case 11: sprintf(text, "( %i ): Es wurde eine Antwort erwartet. Gert stellte gerade selbst eine Anfrage.", error); break;
	default: sprintf(text, "( %i ): Nicht abgefangener Modulfehler!", error);
	}
}


// ##################################################################
/**
 * @brief Ermittlung des Fehlertextes zum Fehlercode fuer Funktionsfehler.
 */
void CChannel4Protocol::GetErrorFunction(short module, short func, short error, char *text)
{
	Trace("%s(%i):INFO(3): CChannel4Protocol::GetErrorFunction [ %i, %i, %i, 0x%08X ]\n", __FILE__, __LINE__, module, func, error, text);

	sprintf(text, "( %i ): Nicht abgefangener Funktionsfehler!", error);
}
