/**
 * @file  IntelHex.h
 * @brief Schnittstelle fuer die Klasse CIntelHex.
 */
//////////////////////////////////////////////////////////////////////
#ifndef _INTELHEX_H_
#define _INTELHEX_H_

#ifdef _WIN32
#pragma warning(disable : 4996)
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

// Grunddaten und Grundoffsets
#define INTELHEX_DATA_COUNT 20  // Anzahl Datenbyte pro Datensatz
#define INTELHEX_START_SIGN ':' // Startzeichen eines Datensatzes
#define INTELHEX_START    0  // start of record
#define INTELHEX_COUNT    1  // byte count
#define INTELHEX_ADDRESS  3  // load address
#define INTELHEX_TYPE     7  // record type
#define INTELHEX_DATA     9  // record data
#define INTELHEX_CHECKSUM 7  // record checksum

// Groessen fuer Textpuffer der Intelhex-Records
#define INTELHEX_LENGTH_HEAD     9 // Anzahl Zeichen fuer Recordbeginn zu Daten
#define INTELHEX_LENGTH_FRAME   (INTELHEX_LENGTH_HEAD + 2) // Anzahl Zeichen fuer Recordbeginn zu Daten + CRC
#define INTELHEX_LENGTH_MAX     (INTELHEX_LENGTH_FRAME + 510) // Anzahl Zeichen fuer Recordframe + Daten

// Standard - Datensatztypen
#define INTELHEX_TYPE_DAT 0x00 // Data record
#define INTELHEX_TYPE_EOF 0x01 // End-of-file record
#define INTELHEX_TYPE_ESA 0x02 // Extended Segment Address Record
#define INTELHEX_TYPE_SSA 0x03 // Start Segment Address Record
#define INTELHEX_TYPE_ELA 0x04 // Extended Linear Address Record
#define INTELHEX_TYPE_SLA 0x05 // Start Linear Address Record

// Datei der Hardwaredaten
#define INTELHEX_TYPE_HARDWARE_HEADER   0x10
#define INTELHEX_TYPE_HARDWARE_EEPROM   0x11
#define INTELHEX_TYPE_HARDWARE_FLASH    0x12

// Datei der Anzeigedaten
#define INTELHEX_TYPE_DISPLAY_HEADER  0x20
#define INTELHEX_TYPE_DISPLAY_FONT    0x21
#define INTELHEX_TYPE_DISPLAY_GRAFIC  0x22

// Datei der Geraetedaten
#define INTELHEX_TYPE_DEVICE_HEADER   0x30
#define INTELHEX_TYPE_DEVICE_EEPROM   0x31
#define INTELHEX_TYPE_DEVICE_FLASH    0x32
#define INTELHEX_TYPE_DEVICE_CRYPT    0x33

// Datei der EVA - Daten
#define INTELHEX_TYPE_DOCUMENT_HEADER    0x40
#define INTELHEX_TYPE_DOCUMENT_EEPROM    0x41
#define INTELHEX_TYPE_DOCUMENT_FLASH     0x42
#define INTELHEX_TYPE_DOCUMENT_XFLASH    0x43
#define INTELHEX_TYPE_DOCUMENT_CONFIG    0x44
#define INTELHEX_TYPE_DOCUMENT_PATH      0x45
#define INTELHEX_TYPE_DOCUMENT_PATH2     0x46

// Datei der Textdaten
#define INTELHEX_TYPE_TEXT_HEADER      0x50
#define INTELHEX_TYPE_TEXT_DATA        0x51

// Datei der Touchkeydaten
#define INTELHEX_TYPE_TOUCHKEY_HEADER  0x60
#define INTELHEX_TYPE_TOUCHKEY_DATA    0x61

// Datei der Displayfarben
#define INTELHEX_TYPE_DISPLAYCOLOR_HEADER 0x70
#define INTELHEX_TYPE_DISPLAYCOLOR_DATA   0x71

// ErrorState der IntelHex-Verarbeitung
#define INTELHEX_ES_NO                      0 // Kein Fehler im Zustand
#define INTELHEX_ES_NO_TEXT                 "Kein Fehler im Zustand"
#define INTELHEX_ES_FILE_OPEN               1 // Datei konnte nicht geoeffnet werden.
#define INTELHEX_ES_FILE_OPEN_TEXT          "Datei konnte nicht geoeffnet werden."
#define INTELHEX_ES_FILE_FORMAT             2 // Dateiinhalt entspricht nicht dem benoetigten Format.
#define INTELHEX_ES_FILE_FORMAT_TEXT        "Dateiinhalt entspricht nicht dem benoetigten Format."
#define INTELHEX_ES_EMPTY                   3 // Die uebergebene Datenzeile ist leer.
#define INTELHEX_ES_EMPTY_TEXT              "Die uebergebene Datenzeile ist leer."
#define INTELHEX_ES_CRC                     4 // Der Berechnete CRC stimmt nicht mit dem im Datensatz ueberein.
#define INTELHEX_ES_CRC_TEXT                "Der Berechnete CRC stimmt nicht mit dem im Datensatz ueberein."
#define INTELHEX_ES_FORMAT_RECORD           5 // Der Datensatz weist fehler auf. Puffer zu kurz, lang oder fehlerhafte Laengenabgabe.
#define INTELHEX_ES_FORMAT_RECORD_TEXT      "Der Datensatz weist fehler auf. Puffer zu kurz, lang oder fehlerhafte Laengenabgabe."
#define INTELHEX_ES_RECORD_OVER_BUFFER      6 // Der Datensatz geht ueber die Grenze des uebergebenen Puffers hinweg.
#define INTELHEX_ES_RECORD_OVER_BUFFER_TEXT "Der Datensatz geht ueber die Grenze des uebergebenen Puffers hinweg."
#define INTELHEX_ES_FILE_ERROR              7 // Es ist ein Dateifehler beim Lesen der Daten aufgetreten.
#define INTELHEX_ES_FILE_ERROR_TEXT         "Es ist ein Dateifehler beim Lesen der Daten aufgetreten."
#define INTELHEX_ES_FILE_EMPTY              8 // In der Datei liegt kein Datensatz vor.
#define INTELHEX_ES_FILE_EMPTY_TEXT         "In der Datei liegt kein Datensatz vor."

class CIntelHex
{
public:
	CIntelHex();
	virtual ~CIntelHex();

// helper
public:
	virtual void Reset();

	void UINT8ToHex(unsigned char val, char *data);
  unsigned char HexToUINT8(const char *data);

	void UINT16ToHex(unsigned short val, char *data);
	unsigned short HexToUINT16(const char *data);

	int MakeCrc(char *data);
	int CheckCrc(const char *data);

public:
	char *PrepareDataLine(char *data, int length);
	int CheckContent(const char *fileName);
	unsigned char GetCount(const char *data);
	unsigned short GetAddress(const char *data);
	unsigned char GetType(const char *data);
	unsigned char GetData(const char *data, unsigned char *dest);
	unsigned char GetData8(const char *data, unsigned char num);
	unsigned short GetData16(const char *data, unsigned char num);
	virtual int MakeTypedDataLine(unsigned short address, unsigned char type, unsigned char *buf, unsigned char length, char *data);
	virtual int MakeDataLine(unsigned short address, unsigned char *buf, unsigned char length, char *data);
	virtual int MakeEndLine(char *data);
	int CheckFile(const char *fileName);

// Fehlerbehandlung
public:
	virtual void ResetError();
	int GetLastErrorState() { return m_errorState; }
	virtual char *GetErrorString(int errorState = -1);
	char *GetErrorText() { return m_errorText; }
	char *GetErrorDetail() { return m_errorDetail; }
protected:
	static int m_errorState;
	static char m_errorString[0x100];
	static char m_errorText[0x100];
	static char m_errorDetail[0x200];
};

#endif

